using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.IO.MemoryMappedFiles;
using System.Threading;
using System.Runtime.InteropServices;

//
// adapted from: https://github.com/joreg/SpoutCSharp
//
// Modified by Lynn Jarvis 20.07.16 - http://spout.zeal.co
//
// SpoutSenderNames
// It seems that a memory map created with CreateOrOpen is not retained.
// if it does not first exist, so another process does not find it.
// But if a new map is created, another process will find it.
// There might be another reason or solution, but this appears to work OK.
//
// Active sender
// If the VVVV sender is the first one, an active sender name memory map is created so
// that it is detected by "SelectSenderPanel" and a receiver shows it straight away
//
//

namespace Spout
{
    public class SpoutSender: IDisposable
    {
        protected const string SenderNamesMMF = "SpoutSenderNames";
        protected const string ActiveSenderMMF = "ActiveSenderName";

        protected const int SpoutWaitTimeout = 100;
        public const int MaxSenders = 10;
        public const int SenderNameLength = 256;

        // Sender memory map
        protected TextureDesc FTextureDesc;
        protected string FSenderName;
        protected MemoryMappedFile FSharedMemory;
        protected MemoryMappedViewStream FSharedMemoryVS;

        // Active sender name memory map
        protected MemoryMappedFile ActiveNameSharedMemory = null;
        protected MemoryMappedViewStream ActiveNameSharedMemoryVS = null;

        // SenderNames shared memory map
        protected MemoryMappedFile SenderNamesSharedMemory = null;
        protected MemoryMappedViewStream SenderNamesSharedMemoryVS = null;
        protected bool SenderNamesCreatedNew = false;

        public SpoutSender(string senderName, uint sharedHandle, uint width, uint height, uint format, uint usage)
        {
            FSenderName = senderName;
            FTextureDesc = new TextureDesc(sharedHandle, width, height, format, usage, new byte[256]);
        }

        // Create an empty SenderNames map if it does not exist yet
        public void CreateSenderNamesMMF()
        {
            var mutex = new Mutex(true, SenderNamesMMF + "_mutex", out SenderNamesCreatedNew);
            if (SenderNamesCreatedNew)
            {
                // Name map does not exist yet, so create a new one
                var len = SenderNameLength * MaxSenders;
                SenderNamesSharedMemory = MemoryMappedFile.CreateNew(SenderNamesMMF, len);
                if (mutex != null) mutex.ReleaseMutex();
                mutex.Dispose();
            }
            else
            {
                SenderNamesCreatedNew = false; // otherwise make sure this flag is false
            }
        }

        // Close the SenderNames map if we created it
        public void CloseSenderNamesMMF()
        {
            if (SenderNamesSharedMemory != null)
                SenderNamesSharedMemory.Dispose();
        }

        public void Dispose()
        {
            // Remove the sender name from the SenderNames memory map
            RemoveNameFromSendersList(FSenderName);

            // Free the sender shared memory map
            if (FSharedMemoryVS != null)
                FSharedMemoryVS.Dispose();
            if (FSharedMemory != null)
                FSharedMemory.Dispose();

            if (SenderNamesCreatedNew)
            {
                // Close the SenderNames map
                CloseSenderNamesMMF();
                // And the active sender name map
                if (ActiveNameSharedMemoryVS != null)
                    ActiveNameSharedMemoryVS.Dispose();
                if (ActiveNameSharedMemory != null)
                    ActiveNameSharedMemory.Dispose();
            }
        }

        public bool Initialize()
        {
            // Check for or create a new SpoutSenderNames memory map
            CreateSenderNamesMMF();

            // Add the sender name to the sendernames memory map
            if (AddNameToSendersList(FSenderName))
            {
                // Create a shared memory map for the sender
                FSharedMemory = MemoryMappedFile.CreateNew(FSenderName, 280);

                // Copy the sender name and description to the map
                FSharedMemoryVS = FSharedMemory.CreateViewStream();
                var nameBytes = Encoding.Unicode.GetBytes(FSenderName);
                Array.Copy(nameBytes, 0, FTextureDesc.Description, 0, nameBytes.Length);
                var desc = FTextureDesc.ToByteArray();
                FSharedMemoryVS.Write(desc, 0, desc.Length);

                // If we are the only sender, create a new active sender name map.
                // This is a separate shared memory containing just a sender name
                // that receivers can use to retrieve the current active Sender.
                // Set our sender as the active Sender if it is the first one registered.
                // Thereafter the user can select an active Sender using SpoutPanel.
                if (SenderNamesCreatedNew)
                {
                    // Create a new active sender map
                    ActiveNameSharedMemory = MemoryMappedFile.CreateNew(ActiveSenderMMF, SenderNameLength);
                    ActiveNameSharedMemoryVS = ActiveNameSharedMemory.CreateViewStream();
                    // Copy the sender name to it
                    var nb = GetNameBytes(FSenderName);
                    // Write the name bytes to the map
                    ActiveNameSharedMemoryVS.Write(nb, 0, nb.Length);
                }

                return true;

            }
        	return false;
        }

        public bool AddNameToSendersList(string name)
        {
            bool createdNew;
            var mutex = new Mutex(true, SenderNamesMMF + "_mutex", out createdNew);
            if (mutex == null)
                return false;
            var success = false;
            try
            {
                if (mutex.WaitOne(SpoutWaitTimeout))
                {
                    success = true;
                }
                else
                {
                    success = false;
                }
            }
            catch (AbandonedMutexException e)
            {
                success = true;    
            }
            finally
            {
                if (success)
                {
                    List<string> senders = GetSenderNames();
                    if (senders.Contains(name))
                    {
                        success = false;
                    }
                    else
                    {
                        senders.Add(name);
                        WriteSenderNamesToMMF(senders);
                    }
                }
                mutex.ReleaseMutex();
                mutex.Dispose();
            }
            return success;
        }

        public void RemoveNameFromSendersList(string name)
        {
            bool createdNew;
            var mutex = new Mutex(true, SenderNamesMMF + "_mutex", out createdNew);
            if (mutex == null)
                return;
            try
            {
                mutex.WaitOne(SpoutWaitTimeout);
            }
            catch (AbandonedMutexException e)
            {
                //Log.Add(e);     
            }
            finally
            {
                var senders = GetSenderNames();
                if (senders.Contains(name))
                {
                    senders.Remove(name);
                    WriteSenderNamesToMMF(senders);
                }
                mutex.ReleaseMutex();
                mutex.Dispose();
            }
        }
        
        // Write all the sender names to the SpoutSenderNames memory map.
        void WriteSenderNamesToMMF(List<string> senders)
        {
            var len = SenderNameLength * MaxSenders;
            MemoryMappedFile mmf = null;

            if (SenderNamesCreatedNew)
            {
                // We created the map
                mmf = MemoryMappedFile.OpenExisting(SenderNamesMMF);
            }
            else
            {
                // The map was created by another process
                mmf = MemoryMappedFile.CreateOrOpen(SenderNamesMMF, len);
            }

            if (mmf != null)
            {
        		using (var mmfVS = mmf.CreateViewStream())
        		{
                    var count = 0;
		            for (int i = 0; i < senders.Count; i++)
		            {
		                var nameBytes = GetNameBytes(senders[i]);
		                mmfVS.Write(nameBytes, 0, nameBytes.Length);
		                count += nameBytes.Length;
		            }
		            var b = new byte[len - count];
		            mmfVS.Write(b, 0, b.Length);

                    // This was removed but it does not make any difference
                    mmfVS.Dispose();
                    mmf.Dispose();

                }
        	}
        }

        public static List<string> GetSenderNames()
        {
        	var namesList = new List<string>();
		    var name = new StringBuilder();
            var len = MaxSenders * SenderNameLength;
            
        	//Read the memory mapped file in to a byte array and close this shit.
        	using (var mmf = MemoryMappedFile.CreateOrOpen(SenderNamesMMF, len))
        	{
        		using (var mmfVS = mmf.CreateViewStream())
        		{
		            var b = new byte[len];
		            mmfVS.Read(b, 0, len);

		            //split into strings searching for the nulls 
		            for (int i = 0; i < len; i++)
		            {
		                if (b[i] == 0)
		                {
		                    if (name.Length == 0)
		                    {
		                        i += SenderNameLength - (i % SenderNameLength) - 1;
		                        continue;
		                    }
		                    namesList.Add(name.ToString());
		                    name.Clear();
		                }
		                else
		                    name.Append((char)b[i]);
		            }
         		}
        	}
        	return namesList;
        }

        protected static byte[] GetNameBytes(string name)
        {
            var b = new byte[SenderNameLength];
            var nameBytes = Encoding.ASCII.GetBytes(name);
            Array.Copy(nameBytes, b, nameBytes.Length);
            return b;
        }
    }
	
	[StructLayout(LayoutKind.Sequential, Pack = 1)]
    public struct TextureDesc
    {
        public uint SharedHandle;
        public uint Width;
        public uint Height;
        public uint Format;
        public uint Usage;
        public byte[] Description;

        public TextureDesc(uint sharedHandle, uint width, uint height, uint format, uint usage, byte[] description)
        {
            SharedHandle = sharedHandle;
            Width = width;
            Height = height;
            Format = format;
            Usage = usage;
            Description = description;
        }

        public TextureDesc(System.IO.MemoryMappedFiles.MemoryMappedViewStream mmvs)
        {
            var br = new BinaryReader(mmvs);
            SharedHandle = br.ReadUInt32();
            Width = br.ReadUInt32();
            Height = br.ReadUInt32();
            Format = br.ReadUInt32();
            Usage = br.ReadUInt32();
            Description = br.ReadBytes(256);
        }

        public byte[] ToByteArray()
        {
            var b = new byte[280];
            var ms = new MemoryStream(b);
            var bw = new BinaryWriter(ms);
            bw.Write(SharedHandle);
            bw.Write(Width);
            bw.Write(Height);
            bw.Write(Format);
            bw.Write(Usage);
            bw.Write(Description,0, Description.Length);
            bw.Dispose();
            ms.Dispose();
            return b;
        }
    }
}
